@@ -2,4 +2,4 @@
 # pylint: disable=unused-import
 # pylint: disable=line-too-long
 
-from utils.parse_sql import drop_table
\ No newline at end of file
+from utils.parse_sql import drop_table,insert_sql
\ No newline at end of file
