@@ -2,4 +2,4 @@
 # pylint: disable=unused-import
 # pylint: disable=line-too-long
 
-from utils.parse_sql import strip_sql_quotes,strip_comments,escape_quoted_chars,parse
+from utils.parse_sql import strip_sql_quotes,strip_comments,escape_quoted_chars,parse,sql_data_types
