@@ -9,6 +9,69 @@ import utils.string_format as format
 import utils.objectUtils as obj
 from pyparsing import And, Suppress,Dict, Word, Literal, Group, Optional, ZeroOrMore, OneOrMore, Regex, restOfLine, alphanums,nums, printables, string, CaselessKeyword,nestedExpr,ParseException,quotedString,removeQuotes,originalTextFor,delimitedList,QuotedString
 
+sql_data_types = [
+    "mediumtext",
+    "mediumblob",
+    "varbinary",
+    "timestamp",
+    "mediumint",
+    "tinytext",
+    "tinyblob",
+    "smallint",
+    "longtext",
+    "longblob",
+    "datetime",
+    "varchar",
+    "tinyint",
+    "integer",
+    "decimal",
+    "boolean",
+    "double",
+    "double",
+    "binary",
+    "bigint",
+    "float",
+    "float",
+    "year",
+    "time",
+    "text",
+    "enum",
+    "date",
+    "char",
+    "bool",
+    "blob",
+    "set",
+    "int",
+    "dec",
+    "bit",
+]
+
+sql_types_with_sizes = [
+    "varbinary",
+    "mediumint",
+    "timestamp",
+    "smallint",
+    "datetime",
+    "varchar",
+    "decimal",
+    "integer",
+    "tinyint",
+    "binary",
+    "double",
+    "double",
+    "bigint",
+    "float",
+    "float",
+    "text",
+    "char",
+    "time",
+    "blob"
+    "dec",
+    "int",
+    "bit",
+]
+
+
 
 def drop_table(table_name,schema_name=None,if_exists=True,quote_char="`"):
     '''
@@ -72,7 +135,297 @@ def drop_table(table_name,schema_name=None,if_exists=True,quote_char="`"):
         exist_string = " IF EXISTS"
     return f'''DROP TABLE{exist_string} {schema_string}`{table_name}`;'''
 
+
+
+def _gen_column_string(columns):
+    '''
+        Generate the SQL list of column names.
+
+        This will not include indices that start with an underscore.
+
+        ----------
+
+        Arguments
+        -------------------------
+        `columns` {list}
+            A list of column name strings.
+
+        Return {str}
+        ----------------------
+        The SQL list of column names.
+
+        Meta
+        ----------
+        `author`: Colemen Atwood
+        `created`: 06-02-2022 09:03:14
+        `memberOf`: TableDataManager
+        `version`: 1.0
+        `method_name`: _gen_column_string
+        * @xxx [06-02-2022 09:04:22]: documentation for _gen_column_string
+    '''
+
+
+    column_array = []
+    # @Mstep [LOOP] iterate the column names.
+    for name in columns:
+        # @Mstep [IF] if the first char is an underscore.
+        if name[0] == "_":
+            # @Mstep [] go to the next indice.
+            continue
+        # @Mstep [] append the column name wrapped in accents to the column_array
+        column_array.append(f"`{name}`")
+    # @Mstep [] join the column_array with commas
+    # @Mstep [RETURN] return column string
+    return ', '.join(column_array)
+
+def _gen_value_string(data,columns):
+    '''
+        Generate the SQL string for each row.
+
+        The columns list is used to force conformity and order for each row.
+
+        ----------
+
+        Arguments
+        -------------------------
+        `data` {list}
+            A list of dictionaries, each representing a row in the table.
+
+        `columns` {list}
+            A list of strings each is a column name.
+
+
+        Return {str}
+        ----------------------
+        The SQL string.
+
+        Meta
+        ----------
+        `author`: Colemen Atwood
+        `created`: 06-02-2022 08:54:10
+        `memberOf`: TableDataManager
+        `version`: 1.0
+        `method_name`: _gen_value_string
+        * @xxx [06-02-2022 08:56:16]: documentation for _gen_value_string
+    '''
+
+
+    value_array = []
+    # @Mstep [LOOP] iterate the column names
+    for col in columns:
+        # @Mstep [] initialize the value as "__NULL_VALUE__"
+        # @Mstep [] this is important because we will need to set the "None" values to literal "null"
+        value = '__NULL_VALUE__'
+
+        # @Mstep [IF] if the column name is a key in the data dictionary
+        if col in data:
+            if data[col] is not None:
+                # @Mstep [] set the value to be the data dictionary's value.
+                value = data[col]
+
+        # @Mstep [IF] if the value from the data dictionary is None
+        if value is None:
+            value = '__NULL_VALUE__'
+
+        # @Mstep [IF] if the value is a string.
+        if isinstance(value,(str)):
+            # @Mstep [] escape single quotes in the value.
+            value = value.replace("'","''")
+            # @Mstep [] append the value to the value_array wrapped in single quotes.
+            value_array.append(f"'{value}'")
+        # @Mstep [ELSE] if the value is another type.
+        else:
+            # @Mstep [] append the value to the value_array
+            value_array.append(f"{value}")
+
+    # for name,value in data.items():
+    #     # @Mstep [IF] if the first char is an underscore
+    #     if name[0] == "_":
+    #         # @Mstep [] skip this column.
+    #         continue
+    #     if value is None:
+    #         value = '__NULL_VALUE__'
+    #     if isinstance(value,(str)):
+    #         value_array.append(f"'{value}'")
+    #     else:
+    #         value_array.append(f"{value}")
+    # @Mstep [] join the value array with commas.
+    value_string = f"({', '.join(value_array)})"
+    # @Mstep [RETURN] return the value_string
+    return value_string
+
+def _get_all_columns(data):
+    '''
+        Get all unique columns from a list of row dictionaries.
+        This makes it so all columns can be represented in the SQL and makes sure
+        that every dictionary does not need every column as a key.
+
+        Keys that being with an underscore are filtered out.
+
+        ----------
+
+        Arguments
+        -------------------------
+        `data` {list}
+            The list of row dictionaries to parse.
+
+        Return {list}
+        ----------------------
+        A list of unique column keys
+
+        Meta
+        ----------
+        `author`: Colemen Atwood
+        `created`: 06-02-2022 08:50:01
+        `memberOf`: TableDataManager
+        `version`: 1.0
+        `method_name`: _get_all_columns
+        * @xxx [06-02-2022 08:52:45]: documentation for _get_all_columns
+    '''
+
+
+    columns = []
+    for row in data:
+        for name,value in row.items():
+            # @Mstep [IF] if the first char is an underscore
+            if name[0] == "_":
+                # @Mstep [] skip this column.
+                continue
+            if name not in columns:
+                columns.append(name)
+    return columns
+    # return columns.sort()
+
+
+
+def insert_sql(data,table_name,schema_name=None):
+    '''
+        Generate an insert SQL statement from a list of data.
+
+        ----------
+
+        Arguments
+        -------------------------
+        `data` {list}
+            A list of dictionaries, each representing a row in the table and each key corresponding
+            to a table column.
+
+            data = [
+                {
+                    `primary_id`:1,
+                    `name`:"kitties and titties",
+                    `timestamp`:None
+                }
+            ]
+
+        `table_name` {str}
+            The name of the table to insert the data into.
+
+        [`schema_name`=None] {str}
+            The optional schema that the table belongs to.
+
+
+        Return {str|bool}
+        ----------------------
+        The SQL insert statement.
+        returns False, if the data provided is not a list or the list is empty
+
+
+        Examples
+        ----------------------
+        data = [
+            {
+                "kitty_id":1,
+                "name":"kitties and titties",
+                "timestamp":123456789,
+            }
+        ]
+        sql = csu.gen.sql.insert_sql(data,'kitties','titties')
+
+        ============================== return ==============================
+
+        INSERT INTO \`titties\`.\`kitties\` (\`kitty_id\`, \`name\`, \`timestamp\`) VALUES
+
+        (1, 'kitties and titties', 123456789);
+
+
+
+        Meta
+        ----------
+        `author`: Colemen Atwood
+        `created`: 06-02-2022 09:13:39
+        `memberOf`: parse_sql
+        `version`: 1.0
+        `method_name`: gen_insert_sql
+        * @TODO []: documentation for gen_insert_sql
+    '''
+
+
+    if isinstance(data,(list)):
+        if len(data) > 0:
+            sql = ''
+            value_array = []
+            # @Mstep [] get a list of all unique column keys from the data.
+            columns = _get_all_columns(data)
+
+            # @Mstep [LOOP] iterate the rows in the data list.
+            for row in data:
+                # @Mstep [] generate the row SQL string and append to the value_array
+                value_array.append(_gen_value_string(row,columns))
+            # @Mstep [] generate the column string.
+            column_string = _gen_column_string(columns)
+            # @Mstep [] join the value_array with commas
+            value_string = ',\n'.join(value_array)
+            # @Mstep [] replace the literal string "__NULL_VALUE__" with null
+            value_string = value_string.replace("'__NULL_VALUE__'",'null')
+            value_string = value_string.replace("__NULL_VALUE__",'null')
+
+            # @Mstep [] generate the schema string, this is for cases where a table has no schema.
+            schema_string = ''
+            # @Mstep [IF] if the schema name was provided.
+            if schema_name is not None:
+                # @Mstep [] generate the schema name string.
+                schema_string = f"`{schema_name}`."
+
+            # @Mstep [] generate the insert statement
+            sql += f"INSERT INTO {schema_string}`{table_name}` ({column_string}) VALUES \n"
+            # @Mstep [] concatenate the value_string onto the insert statement.
+            sql += f"{value_string};"
+
+            # cfu.file.write.write(path,sql)
+            return sql
+    return False
+
+
 def sql_type_to_python(value):
+    '''
+        Attempts to convert an sql type to its python equivalent.
+
+        Keep in mind this is not exact.. at fucking all..
+        This is attempting to predict how the value will need to be stored in the SQL file.
+
+        ----------
+
+        Arguments
+        -------------------------
+        `value` {str}
+            The SQL type to convert.
+
+        Return {list}
+        ----------------------
+        A list of similar types
+
+        Meta
+        ----------
+        `author`: Colemen Atwood
+        `created`: 06-02-2022 07:20:35
+        `memberOf`: parse_sql
+        `version`: 1.0
+        `method_name`: sql_type_to_python
+        * @xxx [06-02-2022 07:22:30]: documentation for sql_type_to_python
+    '''
+
+
     sql_data_types = {
         "mediumtext":["str"],
         "mediumblob":[],
@@ -461,7 +814,7 @@ def parse_table_statement(value=None):
             new_data["constraints"] = d['constraints']
         # print(new_data)
         output = new_data
-        
+
     return output
 
 def _parse_table_column_lines(lines):
